/**
 * Copyright (c) 1993 Richard V. Nash.
 * Copyright (c) 2000 Dan Papasian.  All rights reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef _PLAYERDB_H
#define _PLAYERDB_H

#define PLAYER_VERSION 4 

#define MAX_PLAYER 500
#define MAX_OBSERVE 30		/* max # of games one person can observe */
#define MAX_PLAN 10
#define MAX_FORMULA 9
#define MAX_CENSOR 50
#define MAX_NOTIFY 80
#define MAX_ALIASES 30
#define MAX_SIMUL 100
#define MAX_MESSAGES 40
#define MAX_INCHANNELS 16

#define PLAYER_EMPTY 0
#define PLAYER_NEW 1
#define PLAYER_INQUEUE 2
#define PLAYER_LOGIN 3
#define PLAYER_PASSWORD 4
#define PLAYER_PROMPT 5

#define P_LOGIN 0
#define P_LOGOUT 1

#define SORT_BLITZ 0
#define SORT_STAND 1
#define SORT_ALPHA 2
#define SORT_WILD 3

typedef struct statistics {
  int num, win, los, dra, rating, ltime, best, whenbest;
  double sterr;
} statistics;

typedef struct simul_info_t {
  int numBoards;
  int onBoard;
  /*  int results[MAX_SIMUL]; */  /* replaced by next line. */
  int num_wins, num_draws, num_losses;
  int boards[MAX_SIMUL];
} simul_info_t;

#define PFLAG_REG 0x1
#define PFLAG_OPEN 0x2
#define PFLAG_ROPEN 0x4
#define PFLAG_SIMOPEN 0x8
#define PFLAG_FLIP 0x10
#define PFLAG_ADMINLIGHT 0x20
#define PFLAG_RATED 0x40
#define PFLAG_BLACKSIDE 0x80  /* not done; replacing side. */
#define PFLAG_LASTBLACK 0x100
#define PFLAG_PIN 0x200
#define PFLAG_GIN 0x400
#define PFLAG_AVAIL 0x800 /* to be informed about who is available for games*/
#define PFLAG_PRIVATE 0x1000
#define PFLAG_JPRIVATE 0x2000
#define PFLAG_AUTOMAIL 0x4000
#define PFLAG_MAILMESS 0x8000
#define PFLAG_SHOUT 0x10000
#define PFLAG_CSHOUT 0x20000
#define PFLAG_TELL 0x40000
#define PFLAG_KIBITZ 0x80000
#define PFLAG_NOTIFYBY 0x100000
#define PFLAG_PGN 0x200000
#define PFLAG_BELL 0x400000
#define PFLAG_HIDEINFO 0x800000
#define PFLAG_TOURNEY 0x1000000  /* Not coded yet. */
#define PFLAG_ADS 0x2000000
/* Note: we're starting the last byte, assuming a long int has 4 bytes;
   If we run out, we probably should make a Flag1 and a Flag2. */

#define PFLAG_DEFAULT (PFLAG_OPEN | PFLAG_ROPEN | PFLAG_BELL | PFLAG_SHOUT \
                | PFLAG_CSHOUT | PFLAG_TELL | PFLAG_KIBITZ \
                | PFLAG_ADMINLIGHT)

/* PFLAG_SAVED will make a good mask. */
#define PFLAG_SAVED  (PFLAG_OPEN | PFLAG_ROPEN | PFLAG_RATED | PFLAG_BELL \
                      | PFLAG_PIN | PFLAG_GIN | PFLAG_AVAIL | PFLAG_PRIVATE \
                      | PFLAG_JPRIVATE | PFLAG_AUTOMAIL | PFLAG_MAILMESS \
                      | PFLAG_MAILMESS | PFLAG_SHOUT | PFLAG_CSHOUT \
                      | PFLAG_TELL | PFLAG_KIBITZ | PFLAG_NOTIFYBY | PFLAG_PGN)

typedef struct player {
/* This first block is not saved between logins */
  char *login;
/*  int registered;  */  /* incorporated into Flags. */
  int socket;
  int status;
  int game;
  int opponent; /* Only valid if game is >= 0 */
  int side;     /* Only valid if game is >= 0 */
  int timeOfReg;
  int totalTime;
  int last_tell;
  int ftell; /* Are you forwarding tells?  -1 if not else who from */
  int last_channel;
  int last_opponent;
  int logon_time;
  int last_command_time;
  int num_observe;
  int observe_list[MAX_OBSERVE];
  /*  int lastColor;  */    /* incorporated into Flags. */
  unsigned int thisHost;
/*  timelog *query_log; */    
  int lastshout_a;
  int lastshout_b;
/*  int sopen;  */    /* incorporated into Flags. */
  simul_info_t *simul_info;
  char *busy;
  char *identptr; /* identd stuff */
  int num_comments; /* number of lines in comments file */
/*  int flip;  */    /* incorporated into Flags. */
  List *lists;
  int partner;
  long last_file_byte;
  char *last_file;
/*  int i_admin;  */    /* incorporated into Flags. */
  int kiblevel;
  int number_pend_from; /* not really necessary but are used to cut down */
  int number_pend_to;   /*   search - DAV */

/* Some of this is saved :) */
  unsigned long Flags;

/* All of this is saved between logins */
  char *name;
  char *passwd;
  char *fullName;
  char *emailAddress;
  char *prompt;
  statistics s_stats;
  statistics b_stats;
  statistics w_stats;
  statistics l_stats;
  statistics bug_stats;
  int d_time;
  int d_inc;
  int d_height;
  int d_width;
  int language;
#if 0    /* the following 17 variables have been incorporated into Flags. */
  int open;
  int rated;
  int ropen;
  int notifiedby;
  int bell;
  int pgn;
  int i_login;
  int i_game;
  int i_avail;
  int i_shout;
  int i_cshout;
  int i_tell;
  int i_kibitz;
  int private;
  int jprivate;
  int automail;
  int i_mailmess;
#endif
  int style;
  int promote;
  int adminLevel;
/*  int computer; */  /* 1 if the player is a computer */
  int availmin;
  int availmax;
  int num_plan;
  char *planLines[MAX_PLAN];
  int num_formula;
  char *formulaLines[MAX_FORMULA];
  char *formula;
  int nochannels;
  int num_white;
  int num_black;
  unsigned int lastHost;
  int numAlias;
  alias_type alias_list[MAX_ALIASES];
  int highlight;
} player;

typedef struct textlist {
  char *text;
  int index;
  struct textlist *next;
} textlist;

#define PARRAY_SIZE (MAX_PLAYER + 50)

#endif /* _PLAYERDB_H */

